#include "image.h"
#include "assert.h"
#include "freeimageplus.h"
#include <math.h>
#include "mip.h"

#ifdef DRAW_OPENGL
#include <GL/glew.h>
#endif

void load_into(Array2D<vect4f> &ret, string fn)
{
	// load it from file
	fipImage img;
	img.load(fn.c_str());
	ret.resize(img.getWidth(), img.getHeight());
	int pix_num = ret.s[0] * ret.s[1];

	// upload the data
	if (img.getBitsPerPixel() == 24)
	{
		for (int j = 0; j < ret.s[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < ret.s[0]; i++)
			{
				int idx = j*ret.s[0] + i;
				ret.data[idx][0] = line[i*3 + 2] / 255.0;
				ret.data[idx][1] = line[i*3 + 1] / 255.0;
				ret.data[idx][2] = line[i*3 + 0] / 255.0;
				ret.data[idx][3] = 1;
			}
		}
	}
	else if (img.getBitsPerPixel() == 32)
	{
		for (int j = 0; j < ret.s[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < ret.s[0]; i++)
			{
				int idx = j*ret.s[0] + i;
				ret.data[idx][0] = line[i*4 + 2] / 255.0;
				ret.data[idx][1] = line[i*4 + 1] / 255.0;
				ret.data[idx][2] = line[i*4 + 0] / 255.0;
				ret.data[idx][3] = line[i*4 + 3] / 255.0;
			}
		}
	}
}

void save_image(string fn, Array2D<vect4f> &in, int bits)
{
	// load it from file
	fipImage img;
	img.load(fn.c_str());
	img.setSize(FIT_BITMAP, in.s[0], in.s[1], bits);

	// upload the data
	if (bits == 24)
	{
		for (int j = 0; j < in.s[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < in.s[0]; i++)
			{
				int idx = j*in.s[0] + i;
				line[i*3 + 2] = in.data[idx][0] * 255.0;
				line[i*3 + 1] = in.data[idx][1] * 255.0;
				line[i*3 + 0] = in.data[idx][2] * 255.0;
			}
		}
	}
	else if (bits == 32)
	{
		for (int j = 0; j < in.s[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < in.s[0]; i++)
			{
				int idx = j*in.s[0] + i;
				line[i*3 + 2] = in.data[idx][0] * 255.0;
				line[i*3 + 1] = in.data[idx][1] * 255.0;
				line[i*3 + 0] = in.data[idx][2] * 255.0;
				line[i*3 + 3] = in.data[idx][3] * 255.0;
			}
		}
	}

	img.save(fn.c_str());
}


Image::Image()
{
	init();
}

Image::Image(string fn)
{
	init();
	load(fn);
}

void Image::init()
{
	gl_tex = gl_buf = 0;
}

void Image::load(string fn)
{
#ifdef DRAW_OPENGL
	release();

	// load it from file
	filename = fn;
	fipImage img;
	img.load(fn.c_str());
	size[0] = img.getWidth();
	size[1] = img.getHeight();
	int pix_num = size[0] * size[1];

	if (pix_num == 0)
	{
		printf("ERROR: Failed to load image %s\n", fn.c_str());
	}

	// create buffer
	glGenBuffers(1, &gl_buf);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, gl_buf);

	// upload the data
	if (img.getBitsPerPixel() == 24)
	{
		glBufferData(GL_PIXEL_UNPACK_BUFFER, pix_num*3, 0, GL_STREAM_DRAW);
		vect3ub *ptr = (vect3ub*)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
		assert(ptr);

		for (int j = 0; j < size[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < size[0]; i++)
			{
				int idx = j*size[0] + i;
				ptr[idx][0] = line[i*3 + 2];
				ptr[idx][1] = line[i*3 + 1];
				ptr[idx][2] = line[i*3 + 0];
			}
		}

		glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

		// create square_texture
		glGenTextures(1, &gl_tex);
		glBindTexture(GL_TEXTURE_2D, gl_tex);

		// create square_texture
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, size[0], size[1], 0, GL_RGB, GL_UNSIGNED_BYTE, 0);
	}
	else if (img.getBitsPerPixel() == 32)
	{
		glBufferData(GL_PIXEL_UNPACK_BUFFER, pix_num*4, 0, GL_STREAM_DRAW);
		vect4ub *ptr = (vect4ub*)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
		assert(ptr);

		for (int j = 0; j < size[1]; j++)
		{
			unsigned char *line = img.getScanLine(j);
			for (int i = 0; i < size[0]; i++)
			{
				int idx = j*size[0] + i;
				ptr[idx][0] = line[i*4 + 2];
				ptr[idx][1] = line[i*4 + 1];
				ptr[idx][2] = line[i*4 + 0];
				ptr[idx][3] = line[i*4 + 3];
			}
		}

		glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

		// create square_texture
		glGenTextures(1, &gl_tex);
		glBindTexture(GL_TEXTURE_2D, gl_tex);

		// create square_texture
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, size[0], size[1], 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	}

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
#endif
}


void Image::load(Array2D<vect4f> &img)
{
#ifdef DRAW_OPENGL
	release();

	size[0] = img.s[0];
	size[1] = img.s[1];
	int pix_num = size[0] * size[1];

	// create buffer
	glGenBuffers(1, &gl_buf);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, gl_buf);

	// upload the data
	glBufferData(GL_PIXEL_UNPACK_BUFFER, pix_num*4*4, 0, GL_STREAM_DRAW);
	vect4f *ptr = (vect4f*)glMapBuffer(GL_PIXEL_UNPACK_BUFFER, GL_WRITE_ONLY);
	assert(ptr);

	for (int j = 0; j < img.data_size; j++)
		ptr[j] = img.data[j];

	glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER);

	// create square_texture
	glGenTextures(1, &gl_tex);
	glBindTexture(GL_TEXTURE_2D, gl_tex);

	// create square_texture
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, size[0], size[1], 0, GL_RGBA, GL_FLOAT, 0);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
#endif
}

Image::~Image()
{
	release();
}

void Image::release()
{
#ifdef DRAW_OPENGL
	if (gl_tex)
		glDeleteTextures(1, &gl_tex);
	if (gl_buf)
		glDeleteBuffers(1, &gl_buf);
#endif
}

void Image::bind()
{
#ifdef DRAW_OPENGL
	glBindTexture(GL_TEXTURE_2D, gl_tex);
#endif
}





Image* ImageManager::alloc(string name)
{
	map<string, ImageRef>::iterator it = images.find(name);
	if (it == images.end())
	{
		ImageRef &ref = images[name];
		ref.ptr = new Image(name);
		ref.refs = 1;
		return ref.ptr;
	}
	else
	{
		it->second.refs++;
		if (it->second.refs > 2000000000)
			it->second.refs = 2000000000;
		return it->second.ptr;
	}
}

void ImageManager::free(Image* obj)
{
	if (!obj)
		return;

	// TODO: linear search is slow... do something smarter
	map<string, ImageRef>::iterator it = images.begin();
	for (; it != images.end(); ++it)
	{
		if (it->second.ptr == obj)
		{
			it->second.refs--;
			assert(it->second.refs >= 0);
			if (it->second.refs <= 0)
			{
				delete it->second.ptr;
				images.erase(it);
				return;
			}
		}
	}
}


void Image::draw(vect2f &dst)
{
#ifdef DRAW_OPENGL
	// find texture coordinates
	float texX1 = 0;
	float texY1 = 0;
	float texX2 = 1;
	float texY2 = 1;

	// bind it
	glBindTexture(GL_TEXTURE_2D, gl_tex);

	// draw it
	glBegin(GL_QUADS);
	glTexCoord2f(texX1, texY1);
	glVertex2f(dst[0], dst[1]);

	glTexCoord2f(texX1, texY2);
	glVertex2f(dst[0], dst[1] + size[1]);

	glTexCoord2f(texX2, texY2);
	glVertex2f(dst[0] + size[0], dst[1] + size[1]);

	glTexCoord2f(texX2, texY1);
	glVertex2f(dst[0] + size[0], dst[1]);
	glEnd();
#endif
}
