#include <stdio.h>
#include <time.h>

#include "vect.h"
#include "global.h"
#include "timer.h"
#include "color_space.h"
#include "random.h"

#include <omp.h>

#ifdef DRAW_OPENGL
#include <SDL.h>
#include <GL/glew.h>

#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "glew32.lib")
#endif


#ifdef DRAW_OPENGL
void init_gl()
{
	// init SDL
	putenv("SDL_VIDEO_WINDOW_POS=10,30");

	SDL_Init(SDL_INIT_VIDEO);
	g.screen = SDL_SetVideoMode(g.scr_width, g.scr_height, 32, SDL_HWSURFACE | SDL_OPENGL | SDL_HWACCEL);
	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

	// get version
	const GLubyte *version_gl = glGetString(GL_VERSION);
	printf("OpenGL %s\n", version_gl);
	const GLubyte *version_glsl = glGetString(GL_SHADING_LANGUAGE_VERSION);
	printf("GLSL %s\n", version_glsl);

	// some gl presets
	glewInit();

	glEnable(GL_FRAMEBUFFER_SRGB);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// load fonts
	g.img_font.load("data/font.png");
	g.font.load(&g.img_font);
}

void input()
{
	static bool ignore_next_motion = true;
	static bool grabbed = false;

	if (grabbed)
	{
		SDL_WM_GrabInput(SDL_GRAB_ON);
		SDL_ShowCursor(0);
	}
	else
	{
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		SDL_ShowCursor(1);
	}

	// read states
	Uint8 *keystate = SDL_GetKeyState(0);

	// handle events
	SDL_Event event;
	while (SDL_PollEvent(&event))
	{
		if (event.type == SDL_KEYDOWN)
		{
			if (event.key.keysym.sym == SDLK_ESCAPE)
			{
				grabbed = !grabbed;
				ignore_next_motion = true;
			}
			else if (event.key.keysym.sym >= SDLK_0 && event.key.keysym.sym <= SDLK_9)
			{
				g.mip_level = event.key.keysym.sym - SDLK_0;
			}
			else if (event.key.keysym.sym == SDLK_q)
			{
				g.which = (g.which - 1 + (3*g.mips.s)) % (3*g.mips.s);
			}
			else if (event.key.keysym.sym == SDLK_w)
			{
				g.which = (g.which + 1) % (3*g.mips.s);
			}
			else if (event.key.keysym.sym == SDLK_e)
			{
				g.draw_mesh = !g.draw_mesh;
			}
			else if (event.key.keysym.sym == SDLK_t)
			{
				g.draw_text = !g.draw_text;
			}
			else if (event.key.keysym.sym == SDLK_a)
			{
				g.mip_fraction -= 1.0/16.0;

				if (g.mip_fraction < 0)
				{
					g.mip_fraction++;
					g.mip_level--;
				}
				printf("mip level %f\n", g.mip_level + g.mip_fraction);
			}
			else if (event.key.keysym.sym == SDLK_s)
			{
				g.mip_fraction += 1.0/16.0;
				
				if (g.mip_fraction >= 1)
				{
					g.mip_fraction--;
					g.mip_level++;
				}
				printf("mip level %f\n", g.mip_level + g.mip_fraction);
			}
			else if (event.key.keysym.sym == SDLK_UP)
			{
				g.phi++;
			}
			else if (event.key.keysym.sym == SDLK_DOWN)
			{
				g.phi--;
			}
			else if (event.key.keysym.sym == SDLK_LEFT)
			{
				g.theta++;
			}
			else if (event.key.keysym.sym == SDLK_RIGHT)
			{
				g.theta--;
			}
		}
		else if (event.type == SDL_MOUSEBUTTONDOWN)
		{
		}
		else if (event.type == SDL_MOUSEBUTTONUP)
		{
		}
		else if (event.type == SDL_MOUSEMOTION)
		{

		}
		else if(event.type == SDL_QUIT)
		{
			g.running = false;
		}
	}
}

void draw()
{
	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// bind tex
	glEnable(GL_TEXTURE_2D);

	if (g.which < g.mips.s)
	{
		int which = g.which % g.mips.s;

		int ml = g.mip_level;
		if (ml < 0)
			ml = 0;
		if (ml > g.num_mip_levels-1)
			ml = g.num_mip_levels-1;

		g.mips[which].tex[ml].bind();

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}
	else
	{
		int which = g.which % g.mips.s;

		g.mips[which].stack.bind();

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	}
	
#ifdef MIP_CLAMP
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
#else
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
#endif
	//** draw in object space
	glDisable(GL_BLEND);

	// viewport
	glViewport(0, 0, g.scr_width/2, g.scr_height);

	// orthographic
	if (g.which >= 2*g.mips.s)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(90,1,.005,1000);
	}
	else
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(-.5, .5, -.5, .5, -1, 1);
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	int num_copies = 1;
	if (g.which >= 2*g.mips.s)
	{
		num_copies = 1 << g.mip_level;
	}
	else if (g.which >= g.mips.s)
	{
		num_copies = ceil(pow ( 2.0, (g.mip_level + g.mip_fraction ) ));
	}

	for (int X = -num_copies+1; X < num_copies; X++)
	{
		for (int Y = -num_copies+1; Y < num_copies; Y++)
		{
			glPushMatrix();

			// scale
			if (g.which >= 2*g.mips.s)
			{
				
				double s = .1;
				glScalef(s,s,s);
				glTranslatef(0,-1,0);
				glRotatef(-45,1,0,0);
			}
			else if (g.which >= g.mips.s)
			{
				double s = pow ( 2.0, -(g.mip_level + g.mip_fraction ) );
				glScalef(s,s,s);
			}

			glTranslatef(X, Y, 0);

			if (g.which < 2*g.mips.s)
			{
				glRotatef(-g.phi*15,1,0,0);
				glRotatef(-g.theta*15,0,0,1);
			}

			// texture
			glEnable(GL_TEXTURE_2D);
			glColor3f(1,1,1);
			for (int i = 0; i < g.mesh.tris.s; i++)
			{
				glBegin(GL_TRIANGLES);
				Triangle &t = g.mesh.tris[i];
				for (int j = 0; j < 3; j++)
				{
					vect2f &q = g.mesh.uv[t.t[j]];
					glTexCoord2fv(q.v);

					vect3f &p = g.mesh.pos[t.p[j]];
					glVertex3fv(p.v);
				}
				glEnd();
			}

			// mesh
			if (g.draw_mesh)
			{
				glDisable(GL_TEXTURE_2D);
				glColor3f(1,1,1);
				for (int i = 0; i < g.mesh.tris.s; i++)
				{
					glBegin(GL_LINE_LOOP);
					Triangle &t = g.mesh.tris[i];
					for (int j = 0; j < 3; j++)
					{
						vect3f &p = g.mesh.pos[t.p[j]];
						glVertex3fv(p.v);
					}
					glEnd();
				}
			}

			glPopMatrix();
		}
	}

	//** draw in texture space
	
	// viewport
	glViewport(g.scr_width/2, 0, g.scr_width/2, g.scr_height);

	// orthographic
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, -2, 2);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// texture
	glEnable(GL_TEXTURE_2D);

	glColor3f(1,1,1);
	glBegin(GL_TRIANGLE_FAN);
	glTexCoord2f(0, 0);
	glVertex2f(0, 0);
	glTexCoord2f(1, 0);
	glVertex2f(1, 0);
	glTexCoord2f(1, 1);
	glVertex2f(1, 1);
	glTexCoord2f(0, 1);
	glVertex2f(0, 1);
	glEnd();

	// mesh
	if (g.draw_mesh)
	{
		glDisable(GL_TEXTURE_2D);
		glColor3f(1,1,1);
		for (int i = 0; i < g.mesh.tris.s; i++)
		{
			glBegin(GL_LINE_LOOP);
			vect3i &t = g.mesh.tris[i].t;
			for (int j = 0; j < 3; j++)
			{
				vect2f &p = g.mesh.uv[t[j]];
				glVertex2fv(p.v);
			}
			glEnd();
		}
	}

	//*** HUD
	// viewport
	glViewport(0, 0, g.scr_width, g.scr_height);

	// orthographic
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, g.scr_width, 0, g.scr_height, -1, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// draw name of thing
	if (g.draw_text)
	{
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);

		glPushMatrix();
		glTranslatef(g.scr_width/2, 0, 0);
		g.font.draw(g.mip_names[g.which % g.mips.s]);
		glPopMatrix();
	}
}

void run()
{
	double target_time = get_time();

	while (g.running)
	{
		// interface input and logic
		input();

		// draw
		draw();

		SDL_GL_SwapBuffers();

		// limit framerate
		target_time += 1.0 / 30;
		while (get_time() < target_time)
			; // wait
	}
}

void halt()
{
	printf("Halting program nicely...\n");
	SDL_Quit();
}
#endif

#ifdef CALC_ERROR
vect4f sample_bilinear(Array2D<vect4f> &img, vect2f &pin)
{
	if (img.s[0] == 1 && img.s[1] == 1)
		return img(0,0);

	// get point in image and partial point
	// pin \in [0,1] -> p \in [-.5,s+.5]
	vect2f p;
	vect2i f;
	vect2f t;

	for (int i = 0; i < 2; i++)
	{
		p[i] = pin[i] * img.s[i] - .5;
		f[i] = floor(p[i]);
		t[i] = p[i] - f[i];

#ifdef MIP_CLAMP
		// clamp
		if (p[i] <= 0)
		{
			f[i] = 0;
			t[i] = 0;
		}
		else if (p[i] >= img.s[i]-1)
		{
			f[i] = img.s[i]-2;
			t[i] = 1;
		}
#endif
	}


	// interpolate
#ifdef MIP_CLAMP
	vect4f r;
	r = (1-t[0])*(1-t[1])*img(f[0],f[1]) +
		(1-t[0])*(t[1])*img(f[0],f[1]+1) +
		(t[0])*(1-t[1])*img(f[0]+1,f[1]) +
		(t[0])*(t[1])*img(f[0]+1,f[1]+1);
#else
	vect4f r;
	r = (1-t[0])*(1-t[1])*img(wrap_ret(f[0],img.s[0]),wrap_ret(f[1],img.s[1])) +
		(1-t[0])*(t[1])*img(wrap_ret(f[0],img.s[0]),wrap_ret(f[1]+1,img.s[1])) +
		(t[0])*(1-t[1])*img(wrap_ret(f[0]+1,img.s[0]),wrap_ret(f[1],img.s[1])) +
		(t[0])*(t[1])*img(wrap_ret(f[0]+1,img.s[0]),wrap_ret(f[1]+1,img.s[1]));
#endif

	return r;
}

vect4f sample_trilinear(MipStack &mip, vect2f &pin, float m)
{
	if (m >= mip.img.s-1)
	{
		return mip.img[mip.img.s-1].data[0];
	}
	else if (m < 0)
	{
		return sample_bilinear(mip.img[0], pin);
	}
	else
	{
		int f = floor(m);
		float t = m - f;


		vect4f r;
		r = (1-t)*sample_bilinear(mip.img[f], pin) + (t)*sample_bilinear(mip.img[f+1], pin);
		return r;
	}
}

struct TriArea
{
	float lo, hi;
};

struct TriAreaTable
{
	Array<TriArea> tab;
	float area;

	void add_tri(float a)
	{
		TriArea t;

		if (tab.s)
			t.lo = area;
		else
		{
			t.lo = 0;
			area = 0;
		}

		t.hi = t.lo + a;

		area += a;

		tab.push_back(t);
	}

	int get_idx(float v)
	{
		for (int i = 0; i < tab.s; i++)
		{
			if (tab[i].hi >= v)
				return i;
		}
		assert(false);
		return 0;
	}
};
#endif

void save_mips(MipStack &mip, string name)
{
	for (int j = 0; j < mip.img.s; j++)
	{
		Array2D<vect4f> img;
		img.resize(mip.img[j].s);
		
		// convert from sRGB to linear
		for (int i = 0; i < img.data_size; i++)
		{
			for (int c = 0; c < 3; c++)
				img.data[i][c] = RGB32F_to_sRGB32F(mip.img[j].data[i][c]);
		}

		char buf[1024];
		sprintf(buf, "mip_%d_%s.png", j, name.c_str());
		save_image(buf, img, 24);
	}
}

void create_test(string meshname, string texname, Array<string> &filternames)
{
	// load mesh and texture
	g.mesh.load(meshname);
	g.mesh.scaleToBox();

	load_into(g.img, texname);

	// convert from sRGB to linear
	for (int i = 0; i < g.img.data_size; i++)
	{
		for (int c = 0; c < 3; c++)
			g.img.data[i][c] = sRGB32F_to_RGB32F(g.img.data[i][c]);
	}

	// calc mip levels
	g.num_mip_levels = ceil(log(max(g.img.s[0],g.img.s[1])-.01)/log(2.0)) + 1;
	printf("Image is %dx%d with %d mip levels.\n", g.img.s[0], g.img.s[1], g.num_mip_levels);

	// mipmap with different methods
	Array<double> times;
	double t0, t1;

	g.mips.resize(filternames.s);

	for (int i = 0 ; i < filternames.s; i++)
	{
		t0 = get_time();
		g.mip_names.push_back(filternames[i]);

		if (filternames[i] == "rainbow")
			calc_mip_rainbow(g.mips[i], g.img);
		else if (filternames[i] == "box")
			calc_mip_box_dumb(g.mips[i], g.img);
		else if (filternames[i] == "boxig")
			calc_mip_box_ignore(g.mips[i], g.img, g.mesh);
		else if (filternames[i] == "pambox")
			calc_mip_box_smart(g.mips[i], g.img, g.mesh);
		else if (filternames[i] == "tent")
			calc_mip_tent_dumb(g.mips[i], g.img);
		else if (filternames[i] == "lanczos3")
			calc_mip_lanczos3_dumb(g.mips[i], g.img);
		else if (filternames[i] == "pambi")
			calc_mip_tent_smart(g.mips[i], g.img, g.mesh);
		else if (filternames[i] == "pamtri")
			calc_mip_trilinear_clamp(g.mips[i], g.img, g.mesh);
		else
		{
			printf("Error: The filter named '%s' doesn't exist.\n", filternames[i].c_str());
			exit(0);
		}

		t1 = get_time();
		times.push_back(t1 - t0);
	}


	// save
	for (int j = 0; j < g.mips.s; j++)
		save_mips(g.mips[j], g.mip_names[j]);

	// print times
	printf("Times:\n");
	for (int j = 0; j < g.mips.s; j++)
	{
		printf("%.3f\t%s\n", times[j], g.mip_names[j].c_str());
	}
	printf("\n");

#ifdef CALC_ERROR
	// calculate error
	printf("--== Calculating errors ==--\n");

	Random rand;
	rand.init(10001);

	Array<double> err;
	Array< Array<double> > err_mip;
	err_mip.resize(g.mips.s);
	const int mip_slices = 10;
	const int mip_slices_total = (g.num_mip_levels - 1)*mip_slices + 1;
	for (int j = 0; j < g.mips.s; j++)
	{
		err.push_back(0);
		for (int k = 0; k < mip_slices_total; k++)
			err_mip[j].push_back(0);
	}

#if 1
	// triangle area sampling
	TriAreaTable areas;

	for (int i = 0; i < g.mesh.tris.s; i++)
	{
		float a = g.mesh.get_area(i);
		areas.add_tri(a);
	}
#endif

	int n_samples = 1*1*1000;
	for (int i = 0; i < n_samples; i++)
	{
		vect2f p;
		float m = rand.real3() * (g.num_mip_levels-1);

#if 0
		// uniform sampling
		p[0] = rand.real3();
		p[1] = rand.real3();
#else
		// triangle area sampling
		{
			int tri_idx = areas.get_idx(rand.real3() * areas.area);
			float s = rand.real3();
			float t = rand.real3();
			//reflect
			if (s + t > 1)
			{
				s = 1 - s;
				t = 1 - t;
			}
			vect2f &p0 = g.mesh.uv[g.mesh.tris[tri_idx].t[0]];
			vect2f &p1 = g.mesh.uv[g.mesh.tris[tri_idx].t[1]];
			vect2f &p2 = g.mesh.uv[g.mesh.tris[tri_idx].t[2]];
			p = (p1-p0)*s + (p2-p0)*t + p0;
		}
#endif
		vect4f ref; ref = sample_bilinear(g.img, p);

		for (int j = 0; j < g.mips.s; j++)
		{
			vect4f v; 
			v = sample_trilinear(g.mips[j], p, m);
			
			for (int c = 0; c < 3; c++)
				err[j] += (ref[c] - v[c])*(ref[c] - v[c]);
		
			for (int k = 0; k < mip_slices_total; k++)
			{
				float slice = k * (1.0/mip_slices);
				vect4f v; 
				v = sample_trilinear(g.mips[j], p, slice);

				for (int c = 0; c < 3; c++)
					err_mip[j][k] += (ref[c] - v[c])*(ref[c] - v[c]);
			}
		}
	}

	printf("Trilinear errors:\n");
	for (int j = 0; j < g.mips.s; j++)
	{
		double e = sqrt(err[j] / (3.0*n_samples));
		printf("%16.15f\t%s\n", e, g.mip_names[j].c_str());
	}
	printf("\n");

	// slice errors
	FILE *f = fopen("slices.txt", "wb");
	fprintf(f, "{");
	for (int j = 0; j < g.mips.s; j++)
	{
		if (j)
			fprintf(f, ",{");
		else
			fprintf(f, "{");

		for (int k = 0; k < mip_slices_total; k++)
		{
			if (k)
				fprintf(f, ",");

			double e = sqrt(err_mip[j][k] / (3.0*n_samples));
			fprintf(f, "{%f,%16.15f}", k*(1.0/mip_slices), e);
		}
		fprintf(f, "}");
	}
	fprintf(f, "}");
	fclose(f);
#endif
}


int main(int argc, char **argv)
{
	int nprocs = omp_get_num_procs();
	int nthreads = nprocs;
	
	// copy args into a nice array
	Array<string> args;
	args.resize(argc-1);
	for (int i = 0; i < args.s; i++)
		args[i] = argv[i+1];

	// parse some info from the args (mesh tex [filters])
	string meshname, texname;
	Array<string> filternames;

	if (args.s >= 1)
		meshname = args[0];
	
	if (args.s >= 2)
		texname = args[1];
	
	for (int i = 2; i < args.s; i++)
		filternames.push_back(args[i]);
	
	omp_set_num_threads(nthreads);
	printf("Using %d threads on machine with %d procs\n", nthreads, nprocs);

	// init opengl so that I can load assets
#ifdef DRAW_OPENGL
	init_gl();
#endif
	
	// force a load for my own internal usage
	if (args.s < 2)
	{
#if 1
		printf("usage: pam.exe <model> <texture> [filter]\n");
		printf("\n");
		printf("This utility generates mipmaps using parameterization-aware filters and others.\n");
		printf("The model should be in OBJ format and use only one texture. The texture can be\n");
		printf("in any format supported by FreeImage, which includes PNG, JPG, BMP, TGA. You\n");
		printf("can optionally choose a filter from the list below. If no filter is specified,\n");
		printf("'pambi' is used. Constrained optimization is not included in the binary,\n");
		printf("because we used the commercial library Gurobi.\n");
		printf("\n");
		printf("pamtri       PAM Optimization for trilinear reconstruction\n");
		printf("pambi        PAM Optimization for bilinear reconstruction\n");
		printf("pambox       PAM Optimization for box reconstruction\n");
		printf("lanczos3     Lanczos 3 filter\n");
		printf("tent         Tent filter\n");
		printf("box          Box filter\n");
		printf("boxig        Box filter that ignores unused texels\n");
		printf("rainbow      Each level is given a unique color for testing\n");
		exit(0);
#else
		filternames.push_back("pambi");

		//create_test("simple2.obj", "simple32.png", filternames);
		//create_test("simple2.obj", "boat256.png", filternames);
		//create_test("simple2.obj", "squirrel2.png", filternames);
		//create_test("simple2.obj", "boat256.png", filternames);

		//create_test("brgy2.obj", "brgy5_256.png", filternames);
		
		//create_test("Plasma_Generator.obj", "Plasma_Generator_128.png", filternames);

		create_test("lizard_closed.obj", "lizardmask.png", filternames);

		//create_test("monsterfrog2.obj", "monsterfrog_1024.png", filternames);
#endif
	}
	else
	{
		if (filternames.s == 0)
			filternames.push_back("pambi");

		create_test(meshname, texname, filternames);
	}

#ifdef DRAW_OPENGL
	::run();
	::halt();
#endif

	//for (;;);
	return 0;
}
