#include "mip.h"
#include <float.h>


///////////////////////////////////////////////////////////////////////////////
// Cutting triangles to the texel grid
///////////////////////////////////////////////////////////////////////////////

void split(TexPoly &p, TexPoly &a, TexPoly &b, int dir)
{
	a.uv.s = b.uv.s = 0;
	a.pos.s = b.pos.s = 0;
	const int size = p.uv.s;
	if (!size)
		return;

	int axis1 = 1-dir;

	vect2f *vp = &p.uv[size-1];
	vect3f *posp = &p.pos[size-1];
	double xp = vp->v[dir];
	char inp = xp < 1;

	for (int i = 0; i < size; i++)
	{
		vect2f &v = p.uv[i];
		vect3f &pos = p.pos[i];
		const double &x = v.v[dir];
		const char in = x < 1;

		int sel = inp*2 + in;

		switch (sel)
		{
		case 3:
			//if (inp && in)
			{
				a.uv.push_back(v);
				a.pos.push_back(pos);
			}
			break;
		case 2:
			//else if (inp && !in)
			{
				const double div = x - xp;
				//assert(div > 0); // since the sign is different there must be at least an infinitessimal step difference

				double t = (1 - xp)/div;
				if (t < 0 || t > 1) // only numerical instability can make t be out of range, so assume half because the line is nearly tangent
					t = .5;

				vect2f m;
				float *pp1 = vp->v + axis1;
				m[axis1] = (v[axis1] - *pp1)*t + *pp1;

				m[dir] = 1; // force to be exactly on boundary

				vect3f posm = *posp * (1 - t) + pos * t;

				a.uv.push_back(m);
				a.pos.push_back(posm);

				b.uv.push_back(m);
				b.uv.push_back(v);
				b.pos.push_back(posm);
				b.pos.push_back(pos);
			}
			break;
		case 1:
			//else if (!inp && in)
			{
				const double div = x - xp;
				//assert(div < 0); // since the sign is different there must be at least an infinitessimal step difference

				double t = (1 - xp)/div;
				if (t < 0 || t > 1) // only numerical instability can make t be out of range, so assume half because the line is nearly tangent
					t = .5;

				vect2f m;
				float *pp1 = vp->v + axis1;
				m[axis1] = (v[axis1] - *pp1)*t + *pp1;

				m[dir] = 1; // force to be exactly on boundary

				vect3f posm = *posp * (1 - t) + pos * t;

				a.uv.push_back(m);
				a.uv.push_back(v);
				a.pos.push_back(posm);
				a.pos.push_back(pos);

				b.uv.push_back(m);
				b.pos.push_back(posm);
			}
			break;
		case 0:
			//else
			{
				b.uv.push_back(v);
				b.pos.push_back(pos);
			}
		}

		xp = x;
		vp = &v;
		posp = &pos;
		inp = in;
	}

	for (int i = 0; i < b.uv.s; i++) // yes, it is faster to do subtractions in a loop after verts are added
		b.uv[i].v[dir]--;
}

void add_poly_recursive(Array<Array2D<Texel> > &texels, TexPoly &poly, int x, int y, int depth)
{
	assert(poly.uv.s == poly.pos.s);

	// check for degenaracy
	for (int i = 0; i < poly.uv.s;)
	{
		int ii = (i+1) % poly.uv.s;
		if (poly.pos[i][0] == poly.pos[ii][0] && poly.pos[i][1] == poly.pos[ii][1] && poly.pos[i][2] == poly.pos[ii][2])
		{
			// remove entry i+1
			for (int j = ii; j < poly.uv.s-1; j++)
			{
				poly.uv[j] = poly.uv[j+1];
				poly.pos[j] = poly.pos[j+1];
			}
			poly.uv.pop_back();
			poly.pos.pop_back();
		}
		else
			i++;
	}

	if (poly.uv.s < 3)
		return;

	// add it
	texels[depth](x,y).polys.push_back(poly);

	// terminate
	if (depth == 0)
		return;

	// cut / recur
	for (int i = 0; i < poly.uv.s; i++)
		poly.uv[i] *= 2;

	TexPoly px[2], py[2];
	split(poly, px[0], px[1], 0);

	split(px[0], py[0], py[1], 1);
	if (py[0].uv.s >= 3)
		add_poly_recursive(texels, py[0], x*2 + 0, y*2 + 0, depth-1);
	if (py[1].uv.s >= 3)
		add_poly_recursive(texels, py[1], x*2 + 0, y*2 + 1, depth-1);

	split(px[1], py[0], py[1], 1);
	if (py[0].uv.s >= 3)
		add_poly_recursive(texels, py[0], x*2 + 1, y*2 + 0, depth-1);
	if (py[1].uv.s >= 3)
		add_poly_recursive(texels, py[1], x*2 + 1, y*2 + 1, depth-1);
}

void build_texels(Array<Array2D<Texel> > &texels, Mesh &mesh, int max_mip)
{
	int max_res = (1 << max_mip);
	texels.resize(max_mip+1);
	for (int i = 0; i < texels.s; i++)
		texels[i].resize(1 << (texels.s-i-1), 1 << (texels.s-i-1));

	TexPoly poly;

	for (int i = 0; i < mesh.tris.s; i++)
	{
		vect3f &p0 = mesh.pos[mesh.tris[i].p[0]];
		vect3f &p1 = mesh.pos[mesh.tris[i].p[1]];
		vect3f &p2 = mesh.pos[mesh.tris[i].p[2]];

		vect2f &t0 = mesh.uv[mesh.tris[i].t[0]];
		vect2f &t1 = mesh.uv[mesh.tris[i].t[1]];
		vect2f &t2 = mesh.uv[mesh.tris[i].t[2]];

		poly.uv.resize(3);
		poly.uv[0] = t0;
		poly.uv[1] = t1;
		poly.uv[2] = t2;

		poly.pos.resize(3);
		poly.pos[0] = p0;
		poly.pos[1] = p1;
		poly.pos[2] = p2;

		add_poly_recursive(texels, poly, 0, 0, max_mip);
	}
}

void build_texels(Array2D<Texel> &t, Mesh &mesh, int max_mip)
{
	Array<Array2D<Texel> > texels;
	build_texels(texels, mesh, max_mip);
	t = texels[0];
}




///////////////////////////////////////////////////////////////////////////////
// Upload textures
///////////////////////////////////////////////////////////////////////////////

#ifdef DRAW_OPENGL
#include <GL/glew.h>

void upload_texture(MipStack &mip)
{
	// upload single textures
	for (int i = 1; i < mip.img.s; i++)
	{
		mip.tex[i].load(mip.img[i]);
	}

	// load mip images into first texture
	mip.stack.load(mip.img[0]);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);
	for (int i = 1; i < mip.img.s; i++)
	{
		int w = mip.img[i].s[0];
		int h = mip.img[i].s[1];
		glTexImage2D(GL_TEXTURE_2D, i, GL_RGBA32F, w, h, 0, GL_RGBA, GL_FLOAT, mip.img[i].data);
	}
}
#else
void upload_texture(MipStack &mip)
{
}
#endif



///////////////////////////////////////////////////////////////////////////////
// Utility routines and methods for solving linear systems
///////////////////////////////////////////////////////////////////////////////

double get_area_obj(vect3f &p0, vect3f &p1, vect3f &p2)
{
	vect3f P0 = p1 - p0;
	vect3f P1 = p2 - p0;

	return .5 * (P0%P1).length();
}

double get_area_obj(TexPoly &p)
{
	double sum = 0;
	for (int i = 1; i < p.uv.s - 1; i++)
		sum += get_area_obj(p.pos[0], p.pos[i], p.pos[i+1]);
	return sum;
}


float clamp01(float v)
{
#if 1
	return v;
#else
	if (v < 0)
		return 0;
	else if (v > 1)
		return 1;
	else
		return v;
#endif
}

void fix_unconstrained(SparseTaucsMatrix &A, DenseTaucsMatrix &B)
{
	for (int i = 0; i < A.data.size(); i++)
	{
		if (A.data[i].size() == 0)
		{
			A.data[i][i] = 1;

			for (int j = 0; j < B.cols; j++)
				B(i,j) = 0;
		}
		else
		{
			A.data[i][i] += .001;
		}
	}
}


#ifdef SOLVE_CONSTRAINED
#include <gurobi_c++.h>
#pragma comment(lib, "gurobi46.lib") // gurobi46.lib

void solve_constrained(SparseTaucsMatrix &A, DenseTaucsMatrix &X, DenseTaucsMatrix &B)
{
	X.resize(B.rows, B.cols);

	GRBenv   *env   = NULL;
	GRBmodel *model = NULL;

	Array<int> qrow;
	Array<int> qcol;
	Array<double> qval;

	Array<int> lbeg;
	Array<int> lind;
	Array<double> lval;

	int       optimstatus;
	double    objval;

	/* Create environment */
	GRBloadenv(&env, NULL);

	/* Create an empty model */
	GRBnewmodel(env, &model, NULL, 0, NULL, NULL, NULL, NULL, NULL);

	/* JUST ADD VARIABLES*/
	GRBaddvars(model, B.rows, 
		0, // non-zero terms
		NULL,  // CCS beginnings of cols
		NULL,  // CCS indices of vars
		NULL,  // CCS values
		NULL, // linear term (objective)
		NULL, // lower bound (defaults to 0)
		NULL, // upper bound (? all i want is this, but i guess i have to do it throgh the matrix instead ?)
		NULL, // continuous
		NULL); // names (who cares?)

	/* Integrate new variables */
	GRBupdatemodel(model);

	/* Quadratic objective */
	for (int c = 0; c < A.data.size(); c++)
	{
		for (map<int, double>::iterator it  = A.data[c].begin(); it != A.data[c].end(); ++it)
		{
			int r = it->first;
			double v = it->second;

			qrow.push_back(r);
			qcol.push_back(c);

			if (r == c)
				qval.push_back(v * .5);
			else
				qval.push_back(v);
		}
	}

	GRBaddqpterms(model, qrow.s, qrow.data, qcol.data, qval.data);

	// set constraints
	for (int j = 0; j < B.rows; ++j)
	{
		double val = 1;
		GRBaddconstr(model, 1, &j, &val, GRB_LESS_EQUAL, 1.0, NULL);
	}

	for (int channel = 0; channel < 3; channel++)
	{
		/* set linear part B, (a.k.a objective)*/
		for (int j = 0; j < B.rows; ++j)
		{
			GRBsetdblattrelement(model, "Obj", j, -B(j,channel));
		}

		/* Optimize model */
		GRBoptimize(model);

		/* Capture solution information */
		GRBgetintattr(model, GRB_INT_ATTR_STATUS, &optimstatus);

		GRBgetdblattr(model, GRB_DBL_ATTR_OBJVAL, &objval);

		GRBgetdblattrarray(model, GRB_DBL_ATTR_X, 0, B.rows, X.data + channel*B.rows);

		printf("\nOptimization complete\n");
		if (optimstatus == GRB_OPTIMAL) 
		{
			printf("Optimal objective: %.4e\n", objval);
		} 
		else if (optimstatus == GRB_INF_OR_UNBD) 
		{
			printf("Model is infeasible or unbounded\n");
		} 
		else 
		{
			printf("Optimization was stopped early\n");
		}
	}

	/* Free model */
	GRBfreemodel(model);

	/* Free environment */
	GRBfreeenv(env);
}
#endif
